/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_protocol.h"
#include "chirp_client.h"
#include "chirp_group.h"

#include "sleeptools.h"

#include "link.h"
#include "auth.h"
#include "auth_hostname.h"
#include "domain_name_cache.h"
#include "full_io.h"
#include "macros.h"
#include "debug.h"
#include "copy_stream.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/stat.h>
#ifdef HAS_SYS_STATFS_H
#include <sys/statfs.h>
#endif
#include <sys/param.h>
#include <sys/mount.h>
#include <signal.h>

#include <unistd.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

/* The maximum chunk of memory the server will allocate to handle I/O */
#define MAX_BUFFER_SIZE (16*1024*1024)

static INT64_T get_stat_result( struct chirp_client *c, struct chirp_stat *buf, time_t stoptime );
static INT64_T get_statfs_result( struct chirp_client *c, struct chirp_statfs *buf, time_t stoptime );
static INT64_T get_result( struct chirp_client *c, time_t stoptime );
static INT64_T convert_result( INT64_T response );
static INT64_T simple_command(struct chirp_client *c, time_t stoptime, char const *fmt,...);

static int global_serial = 0;

struct chirp_client {
	struct link *link;
	char hostport[CHIRP_PATH_MAX];
	int broken;
	int serial;
};

struct chirp_client * chirp_client_connect_condor( time_t stoptime )
{
	FILE *file;
	int fields;
	int save_errno;
	struct chirp_client *client;
	char host[CHIRP_LINE_MAX];
	char hostport[CHIRP_LINE_MAX];
	char cookie[CHIRP_LINE_MAX];
	int port;
	int result;

	file = fopen("chirp.config","r");
	if(!file) return 0;

	fields = fscanf(file,"%s %d %s",host,&port,cookie);
	fclose(file);

	if(fields!=3) {
		errno = EINVAL;
		return 0;
	}

	sprintf(hostport,"%s:%d",host,port);

	client = chirp_client_connect(hostport,0,stoptime);
	if(!client) return 0;

	result = chirp_client_cookie(client,cookie,stoptime);
	if(result!=0) {
		save_errno = errno;
		chirp_client_disconnect(client);
		errno = save_errno;
		return 0;
	}

	return client;
}

struct chirp_client * chirp_client_connect( const char *hostport, int negotiate_auth, time_t stoptime )
{
	struct chirp_client *c;
	char addr[LINK_ADDRESS_MAX];
	char host[DOMAIN_NAME_MAX];
	int save_errno;
	int port;

	if(sscanf(hostport,"%[^:]:%d",host,&port)==2) {
		/* use the split host and port */
	} else {
		strcpy(host,hostport);
		port = CHIRP_PORT;
	}

	if(!domain_name_cache_lookup(host,addr)) {
		errno = ENOENT;
		return 0;
	}

	c = malloc(sizeof(*c));
	if(c) {
	        c->link = link_connect(addr,port,stoptime);
		c->broken = 0;
		c->serial = global_serial++;
		if(c->link) {
		        link_tune(c->link,LINK_TUNE_INTERACTIVE);
			if(negotiate_auth) {
				char *type, *subject;
				if(auth_assert(c->link,&type,&subject,stoptime)) {
					free(type);
					free(subject);
					strcpy(c->hostport,hostport);

					return c;
				} else {
					chirp_client_disconnect(c);
					c = 0;
					if(time(0)>=stoptime) {
						errno = ECONNRESET;
					} else {
						errno = EACCES;
					}
				}
			} else {
				return c;
			}
		}
		save_errno = errno;
		free(c);
		errno = save_errno;
	}

	return 0;
}

void chirp_client_disconnect( struct chirp_client *c )
{
	link_close(c->link);
	free(c);
}

static const char * sanitize( const char *s )
{
	static char newpath[CHIRP_PATH_MAX];
	char *n = newpath;
	while(*s) {
		if(isspace(*s) || !isprint(*s)) {
			*n = '_';
		} else {
			*n = *s;
		}
		s++;
		n++;
	}		

	*n = 0;
	return newpath;
}

static const char * sanitize2( const char *s )
{
	static char newpath[CHIRP_PATH_MAX];
	char *n = newpath;
	while(*s) {
		if(isspace(*s) || !isprint(*s)) {
			*n = '_';
		} else {
			*n = *s;
		}
		s++;
		n++;
	}		

	*n = 0;
	return newpath;
}

INT64_T chirp_client_serial( struct chirp_client *c )
{
	return c->serial;
}

INT64_T chirp_client_cookie( struct chirp_client *c, const char *cookie, time_t stoptime )
{
	return simple_command(c,stoptime,"cookie %s\n",cookie);
}

INT64_T chirp_client_login( struct chirp_client *c, const char *name, const char *password, time_t stoptime  )
{
	return simple_command(c,stoptime,"login %s %s\n",sanitize(name),password);
}

INT64_T chirp_client_lookup( struct chirp_client *c, const char *logical_name, char **url, time_t stoptime  )
{
	INT64_T result;
	INT64_T actual;

	result = simple_command(c,stoptime,"lookup %s\n",sanitize(logical_name));

	if(result>0) {
		*url = malloc(result);
		if(*url) {
			actual = link_read(c->link,*url,result,stoptime);
			if(actual!=result) {
				c->broken = 1;
				errno = ECONNRESET;
				return -1;
			}
		} else {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

INT64_T chirp_client_getdir( struct chirp_client *c, const char *path, chirp_dir_t callback, void *arg, time_t stoptime )
{
	INT64_T result;
	const char *name;

	result = chirp_client_opendir(c,sanitize(path),stoptime);
	if(result==0) {
		while((name=chirp_client_readdir(c,stoptime))) {
			callback(name,arg);
		}
	}

	return result;
}

INT64_T chirp_client_opendir( struct chirp_client *c, const char *path, time_t stoptime  )
{
	return simple_command(c,stoptime,"getdir %s\n",sanitize(path));
}

const char * chirp_client_readdir( struct chirp_client *c, time_t stoptime )
{
	static char name[CHIRP_PATH_MAX];

	if(link_readline(c->link,name,sizeof(name),stoptime)) {
		if(name[0]) {
			return name;
		} else {
			return 0;
		}
	} else {
		c->broken = 1;
		errno = ECONNRESET;
		return 0;
	}

}

INT64_T chirp_client_getacl( struct chirp_client *c, const char *path, chirp_dir_t callback, void *arg, time_t stoptime )
{
	INT64_T result;
	const char *name;

	result = chirp_client_openacl(c,sanitize(path),stoptime);
	if(result==0) {
		while((name=chirp_client_readacl(c,stoptime))) {
			callback(name,arg);
		}
	}

	return result;
}

INT64_T chirp_client_openacl( struct chirp_client *c, const char *path, time_t stoptime  )
{
	return simple_command(c,stoptime,"getacl %s\n",sanitize(path));
}

const char * chirp_client_readacl( struct chirp_client *c, time_t stoptime )
{
	static char acl[CHIRP_PATH_MAX];

	if(link_readline(c->link,acl,sizeof(acl),stoptime)) {
		if(acl[0]) {
			return acl;
		} else {
			return 0;
		}
	} else {
		c->broken = 1;
		errno = ECONNRESET;
		return 0;
	}

}

INT64_T chirp_client_setacl( struct chirp_client *c, const char *path, const char *user, const char *acl, time_t stoptime )
{
	return simple_command(c,stoptime,"setacl %s %s %s\n",sanitize(path),sanitize2(user),acl);
}

INT64_T chirp_client_open( struct chirp_client *c, const char *path, INT64_T flags, INT64_T mode, struct chirp_stat *buf, time_t stoptime  )
{
	INT64_T result;
	char fstr[256];
	fstr[0] = 0;

	if(flags&O_WRONLY) {
		strcat(fstr,"w");
	} else if(flags&O_RDWR) {
		strcat(fstr,"rw");
	} else {
		strcat(fstr,"r");
	}

	if(flags&O_CREAT) strcat(fstr,"c");
	if(flags&O_TRUNC) strcat(fstr,"t");
	if(flags&O_APPEND) strcat(fstr,"a");
#ifdef O_SYNC
	if(flags&O_SYNC) strcat(fstr,"s");
#endif
	
	result = simple_command(c,stoptime,"open %s %s %lld\n",sanitize(path),fstr,mode);
	if(result>=0) {
		if(get_stat_result(c,buf,stoptime)>=0) {
			return result;
		} else {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	} else {
		return result;
	}
}

INT64_T chirp_client_close( struct chirp_client *c, INT64_T fd, time_t stoptime  )
{
	return simple_command(c,stoptime,"close %lld\n",fd);
}

INT64_T chirp_client_pread( struct chirp_client *c, INT64_T fd, void *buffer, INT64_T length, INT64_T offset, time_t stoptime  )
{
	INT64_T result;
	INT64_T actual;

	result = simple_command(c,stoptime,"pread %lld %lld %lld\n",fd,length,offset);

	if( result>0 ) {
		actual = link_read(c->link,buffer,result,stoptime);
		if(actual!=result) {
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

INT64_T chirp_client_getfile( struct chirp_client *c, const char *path, FILE *stream, time_t stoptime  )
{
	INT64_T length;

	length = simple_command(c,stoptime,"getfile %s\n",sanitize(path));

	if( length>=0 ) {
		if(link_stream_to_file(c->link,stream,length,stoptime)==length) {
			return length;
		} else {
			c->broken = 1;
			errno = ECONNRESET;
		}
	}
	return -1;
}

INT64_T chirp_client_getfile_buffer( struct chirp_client *c, const char *path, char **buffer, time_t stoptime  )
{
	INT64_T length;
	INT64_T result;

	*buffer = 0;

	length = simple_command(c,stoptime,"getfile %s\n",sanitize(path));
	if(length<=0) return length;

	*buffer = malloc(length+1);
	if(!*buffer) {
		c->broken = 1;
		errno = ENOMEM;
		return -1;
	}

	result = link_read(c->link,*buffer,length,stoptime);
	if(result<0) {
		free(*buffer);
		c->broken = 1;
		return -1;
	}

	(*buffer)[length] = 0;

	return result;
}

INT64_T chirp_client_readlink( struct chirp_client *c, const char *path, char *buffer, INT64_T length, time_t stoptime  )
{
	INT64_T result;
	INT64_T actual;

	result = simple_command(c,stoptime,"readlink %s %lld\n",sanitize(path),length);

	if( result>0 ) {
		actual = link_read(c->link,buffer,result,stoptime);
		if(actual!=result) {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

INT64_T chirp_client_whoami( struct chirp_client *c, char *buffer, INT64_T length, time_t stoptime  )
{
	INT64_T result;
	INT64_T actual;

	result = simple_command(c,stoptime,"whoami %lld\n",length);

	if( result>0 ) {
		actual = link_read(c->link,buffer,result,stoptime);
		if(actual!=result) {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
		buffer[actual] = 0;
	}

	return result;
}

INT64_T chirp_client_whoareyou( struct chirp_client *c, const char *rhost, char *buffer, INT64_T length, time_t stoptime )
{
	INT64_T result;
	INT64_T actual;

	result = simple_command(c,stoptime,"whoareyou %s %lld\n",rhost,length);

	if( result>0 ) {
		actual = link_read(c->link,buffer,result,stoptime);
		if(actual!=result) {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

INT64_T chirp_client_pwrite( struct chirp_client *c, INT64_T fd, const void *buffer, INT64_T length, INT64_T offset, time_t stoptime  )
{
	char line[CHIRP_LINE_MAX];
	INT64_T result;

	//chop off crazy write lengths
	if( length > MAX_BUFFER_SIZE )
		length = MAX_BUFFER_SIZE;

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	//set up command to be sent to server
	sprintf(line,"pwrite %lld %lld %lld\n",fd,length,offset);
	debug(D_CHIRP,"%s: %s",c->hostport,line);

	//write command to link
	result = link_write(c->link,line,strlen(line),stoptime);

	//check to see if command went through, else error in connection
	if(result<0) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	//write the data onto the link
	result = link_write(c->link,buffer,length,stoptime);
	//make sure all the data went out, if not error in connection
	if(result!=length) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	//get how much the server actually wrote to disk/error code
	result = get_result(c,stoptime);
	debug(D_CHIRP,"= %lld",result);
	return convert_result(result);
}

INT64_T chirp_client_putfile( struct chirp_client *c, const char *path, FILE *stream, INT64_T mode, INT64_T length, time_t stoptime )
{

	char line[CHIRP_LINE_MAX];
	INT64_T result;

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	sprintf(line,"putfile %s %lld %lld\n",sanitize(path),mode,length);
	debug(D_CHIRP,"%s: %s",c->hostport,line);

	result = link_write(c->link,line,strlen(line),stoptime);
	if(result<0) goto failure;

	result = get_result(c,stoptime);
	debug(D_CHIRP,"= %lld",result);
	if(result<0) return convert_result(result);

	result = link_stream_from_file(c->link,stream,length,stoptime);
	if(result!=length) goto failure;
	
	return convert_result(get_result(c,stoptime));

failure:
	c->broken = 1;
	errno = ECONNRESET;
	return -1;
}

INT64_T chirp_client_putfile_buffer( struct chirp_client *c, const char *path, const char *buffer, INT64_T mode, INT64_T length, time_t stoptime )
{

	char line[CHIRP_LINE_MAX];
	INT64_T result;

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	sprintf(line,"putfile %s %lld %lld\n",sanitize(path),mode,length);
	debug(D_CHIRP,"%s: %s",c->hostport,line);

	result = link_write(c->link,line,strlen(line),stoptime);
	if(result<0) goto failure;

	result = get_result(c,stoptime);
	debug(D_CHIRP,"= %lld",result);
	if(result<0) return convert_result(result);

	result = link_write(c->link,buffer,length,stoptime);
	if(result!=length) goto failure;
	
	return convert_result(get_result(c,stoptime));

failure:
	c->broken = 1;
	errno = ECONNRESET;
	return -1;
}

INT64_T chirp_client_thirdput( struct chirp_client *c, const char *path, const char *hostname, const char *newpath, time_t stoptime ) {
	return simple_command( c, stoptime, "thirdput %s %s %s\n", sanitize(path), hostname, sanitize2(newpath) );
}

INT64_T chirp_client_fchmod( struct chirp_client *c, INT64_T fd, INT64_T mode, time_t stoptime )
{
	return simple_command(c,stoptime,"fchmod %lld %lld\n",fd,mode);
}

INT64_T chirp_client_fchown( struct chirp_client *c, INT64_T fd, INT64_T uid, INT64_T gid, time_t stoptime )
{
	return simple_command(c,stoptime,"fchown %lld %lld %lld\n",fd,uid,gid);
}

INT64_T chirp_client_ftruncate( struct chirp_client *c, INT64_T fd, INT64_T length, time_t stoptime )
{
	return simple_command(c,stoptime,"ftruncate %lld %lld\n",fd,length);
}

INT64_T chirp_client_fstat( struct chirp_client *c, INT64_T fd, struct chirp_stat *buf, time_t stoptime )
{
	INT64_T result = simple_command(c,stoptime,"fstat %lld\n",fd);
	if(result>=0) result = get_stat_result(c,buf,stoptime);
	return result;
}

INT64_T chirp_client_stat( struct chirp_client *c, const char *path, struct chirp_stat *buf, time_t stoptime )
{
	INT64_T result = simple_command(c,stoptime,"stat %s\n",sanitize(path));
	if(result>=0) result = get_stat_result(c,buf,stoptime);
	return result;
}

INT64_T chirp_client_lstat( struct chirp_client *c, const char *path, struct chirp_stat *buf, time_t stoptime )
{
	INT64_T result = simple_command(c,stoptime,"lstat %s\n",sanitize(path));
	if(result>=0) result = get_stat_result(c,buf,stoptime);
	return result;
}

INT64_T chirp_client_fstatfs( struct chirp_client *c, INT64_T fd, struct chirp_statfs *buf, time_t stoptime )
{
	INT64_T result = simple_command(c,stoptime,"fstatfs %lld\n",fd);
	if(result>=0) result = get_statfs_result(c,buf,stoptime);
	return result;
}

INT64_T chirp_client_statfs( struct chirp_client *c, const char *path, struct chirp_statfs *buf, time_t stoptime )
{
	INT64_T result = simple_command(c,stoptime,"statfs %s\n",sanitize(path));
	if(result>=0) result = get_statfs_result(c,buf,stoptime);
	return result;
}

INT64_T chirp_client_unlink( struct chirp_client *c, const char *path, time_t stoptime )
{
	return simple_command(c,stoptime,"unlink %s\n",sanitize(path));
}

INT64_T chirp_client_rename( struct chirp_client *c, const char *oldpath, const char *newpath, time_t stoptime )
{
	return simple_command(c,stoptime,"rename %s %s\n",sanitize(oldpath),sanitize2(newpath));
}

INT64_T chirp_client_link( struct chirp_client *c, const char *oldpath, const char *newpath, time_t stoptime )
{
	return simple_command(c,stoptime,"link %s %s\n",sanitize(oldpath),sanitize2(newpath));
}

INT64_T chirp_client_symlink( struct chirp_client *c, const char *oldpath, const char *newpath, time_t stoptime )
{
	return simple_command(c,stoptime,"symlink %s %s\n",sanitize(oldpath),sanitize2(newpath));
}

INT64_T chirp_client_fsync( struct chirp_client *c, INT64_T fd, time_t stoptime )
{
	return simple_command(c,stoptime,"fsync %lld\n",fd);
}

INT64_T chirp_client_mkdir( struct chirp_client *c, char const *path, INT64_T mode, time_t stoptime )
{
	return simple_command(c,stoptime,"mkdir %s %lld\n",sanitize(path),mode);
}

INT64_T chirp_client_rmdir( struct chirp_client *c, char const *path, time_t stoptime )
{
	return simple_command(c,stoptime,"rmdir %s\n",sanitize(path));
}

INT64_T chirp_client_truncate( struct chirp_client *c, char const *path, INT64_T length, time_t stoptime )
{
	return simple_command(c,stoptime,"truncate %s %lld\n",sanitize(path),length);
}

INT64_T chirp_client_utime( struct chirp_client *c, char const *path, time_t actime, time_t modtime, time_t stoptime )
{
	return simple_command(c,stoptime,"utime %s %u %u\n",sanitize(path),actime,modtime);
}

INT64_T chirp_client_access( struct chirp_client *c, char const *path, INT64_T mode, time_t stoptime )
{
	return simple_command(c,stoptime,"access %s %lld\n",sanitize(path),mode);
}

INT64_T chirp_client_chmod( struct chirp_client *c, char const *path, INT64_T mode, time_t stoptime )
{
	return simple_command(c,stoptime,"chmod %s %lld\n",sanitize(path),mode);
}

INT64_T chirp_client_chown( struct chirp_client *c, char const *path, INT64_T uid, INT64_T gid, time_t stoptime)
{
	return simple_command(c,stoptime,"chown %s %lld %lld\n",sanitize(path),uid,gid);
}

INT64_T chirp_client_lchown( struct chirp_client *c, char const *path, INT64_T uid, INT64_T gid, time_t stoptime )
{
	return simple_command(c,stoptime,"lchown %s %lld %lld\n",sanitize(path),uid,gid);
}

INT64_T chirp_client_exec( struct chirp_client *c, const char *infile, const char *outfile, const char *errfile, const char *cmd, time_t stoptime )
{
	return simple_command(c,stoptime,"exec %s %s %s %s\n",infile,outfile,errfile,cmd);
}

INT64_T chirp_client_md5( struct chirp_client *c, const char *path, unsigned char digest[16], time_t stoptime  )
{
	INT64_T result;
	INT64_T actual;

	result = simple_command(c,stoptime,"md5 %s\n",path);

	if( result==16 ) {
		actual = link_read(c->link,(char *)digest,16,stoptime);
		if(actual!=result) {
			errno = ECONNRESET;
			result = -1;
		}
		
	} else if(result>=0) {
		result = -1;
		errno = ECONNRESET;
	}
	return result;
}

INT64_T chirp_client_audit( struct chirp_client *c, const char *path, struct chirp_audit **list, time_t stoptime )
{	
	INT64_T result;
	struct chirp_audit *entry;
	int i, actual;
	char line[CHIRP_LINE_MAX];

	result = simple_command(c,stoptime,"audit %s\n",sanitize(path));
	if(result<=0) return result;

	*list = malloc(sizeof(struct chirp_audit)*result);
	entry = *list;

	for(i=0;i<result;i++) {
		actual = link_readline(c->link,line,sizeof(line),stoptime);
		if(actual<=0) {
			free(*list);
			result = -1;
			errno = ECONNRESET;
			break;
		} else {
			sscanf(line,"%s %lld %lld %lld",entry->name,&entry->nfiles,&entry->ndirs,&entry->nbytes);
		}
		entry++;
	}

	return result;
}

INT64_T chirp_client_mkalloc( struct chirp_client *c, char const *path, INT64_T size, INT64_T mode, time_t stoptime )
{
	return simple_command(c,stoptime,"mkalloc %s %lld %lld\n",sanitize(path),size,mode);
}

INT64_T chirp_client_lsalloc( struct chirp_client *c, char const *path, char *allocpath, INT64_T *total, INT64_T *inuse, time_t stoptime )
{
	int result;
	char line[CHIRP_LINE_MAX];

	result = simple_command(c,stoptime,"lsalloc %s\n",sanitize(path));
	if(result==0) {
		if(link_readline(c->link,line,sizeof(line),stoptime)) {
			sscanf(line,"%s %lld %lld",allocpath,total,inuse);
		} else {
			c->broken = 1;
			errno = ECONNRESET;
			result = -1;
		}
	}
	return result;
}

INT64_T chirp_client_creategroup( struct chirp_client *c, char* grp_name, time_t stoptime )
{
        return simple_command(c, stoptime, "creategroup %s\n", grp_name);
}

INT64_T chirp_client_adduser( struct chirp_client *c, char* path, char* usr_name, time_t stoptime )
{
        return simple_command(c, stoptime, "adduser %s %s\n", path, usr_name);
}

INT64_T chirp_client_removeuser( struct chirp_client *c, char* path, char* usr_name, time_t stoptime )
{
        return simple_command(c, stoptime, "removeuser %s %s\n", path, usr_name);
}

INT64_T chirp_client_open_group( struct chirp_client *c, char* path, time_t stoptime )
{
        return simple_command(c, stoptime, "membership %s\n", path);
}

INT64_T chirp_client_set_policy( struct chirp_client *c, char* path, unsigned long int file_duration, unsigned long int dec_duration, time_t stoptime )
{
        return simple_command(c, stoptime, "setpolicy %s %ld %ld\n", path, file_duration, dec_duration);
}

INT64_T chirp_client_update_cache( struct chirp_client* c, const char* path, time_t mod_time, time_t stoptime )
{
        return simple_command(c, stoptime, "check_cache_file_exp %s %ld\n", path, mod_time);
}

INT64_T chirp_client_group_lookup( struct chirp_client* c, const char* grp_name, const char* usr_name, time_t stoptime )
{
       return simple_command(c, stoptime, "lookup %s %s\n", grp_name, usr_name);
}

char * chirp_client_get_group_members( struct chirp_client *c, time_t stoptime )
{
  static char name[CHIRP_PATH_MAX];

  if(link_readline(c->link,name,sizeof(name),stoptime)) {
    if(name[0]) {
      return name;
    } else {
      get_result(c, stoptime);
      return NULL;
    }
  } else {
    c->broken = 1;
    errno = ECONNRESET;
    return NULL;
  }
}

INT64_T chirp_client_get_remote_policy( struct chirp_client* c, const char* grp_name, int* policy, int* file_duration, int* dec_duration, time_t stoptime )
{
  char line[CHIRP_LINE_MAX];
  INT64_T result;

  if(c->broken) {
    errno = ECONNRESET;
    return -1;
  }

  sprintf(line, "get_cache_policy %s\n", grp_name);
  result = link_write(c->link,line,strlen(line),stoptime);
  debug(D_CHIRP,"= %lld",result);
 
  link_readline(c->link,line,CHIRP_LINE_MAX,stoptime);
  result = get_result(c,stoptime);
  sscanf(line,"%d %d %d", policy, file_duration, dec_duration);

  return result;
}

static INT64_T convert_result( INT64_T result )
{
	if(result>=0) {
		return result;
	} else {
		switch(result) {
		case CHIRP_ERROR_NOT_AUTHENTICATED:
		case CHIRP_ERROR_NOT_AUTHORIZED:
			errno = EACCES;
			break;
		case CHIRP_ERROR_DOESNT_EXIST:
			errno = ENOENT;
			break;
		case CHIRP_ERROR_ALREADY_EXISTS:
			errno = EEXIST;
			break;
		case CHIRP_ERROR_TOO_BIG:
			errno = EFBIG;
			break;
		case CHIRP_ERROR_NO_SPACE:
			errno = ENOSPC;
			break;
		case CHIRP_ERROR_NO_MEMORY:
			errno = ENOMEM;
			break;
		case CHIRP_ERROR_INVALID_REQUEST:
			errno = EINVAL;
			break;
		case CHIRP_ERROR_TOO_MANY_OPEN:
			errno = EMFILE;
			break;
		case CHIRP_ERROR_BUSY:
			errno = EBUSY;
			break;
		case CHIRP_ERROR_TRY_AGAIN:
			errno = EAGAIN;
			break;
		case CHIRP_ERROR_NOT_DIR:
			errno = ENOTDIR;
			break;
		case CHIRP_ERROR_IS_DIR:
			errno = EISDIR;
			break;
		case CHIRP_ERROR_NOT_EMPTY:
			errno = ENOTEMPTY;
			break;
		case CHIRP_ERROR_CROSS_DEVICE_LINK:
			errno = EXDEV;
			break;
		case CHIRP_ERROR_GRP_UNREACHABLE:
		case CHIRP_ERROR_TIMED_OUT:
		case CHIRP_ERROR_DISCONNECTED:
		case CHIRP_ERROR_UNKNOWN:
			errno = ECONNRESET;
			break;
		}
		return -1;
	}
}

static INT64_T get_stat_result( struct chirp_client *c, struct chirp_stat *buf, time_t stoptime )
{
	char line[CHIRP_LINE_MAX];
	INT64_T fields;

	memset(buf,0,sizeof(*buf));

	if(!link_readline(c->link,line,CHIRP_LINE_MAX,stoptime)) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	fields = sscanf(line,"%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %llu %llu %llu\n",
			&buf->cst_dev,
			&buf->cst_ino,
			&buf->cst_mode,
			&buf->cst_nlink,
			&buf->cst_uid,
			&buf->cst_gid,
			&buf->cst_rdev,
			&buf->cst_size,
			&buf->cst_blksize,
			&buf->cst_blocks,
			&buf->cst_atime,
			&buf->cst_mtime,
			&buf->cst_ctime);

	buf->cst_dev = -1;
	buf->cst_rdev = 0;

	if(fields!=13) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	return 0;
}

static INT64_T get_statfs_result( struct chirp_client *c, struct chirp_statfs *buf, time_t stoptime )
{
	char line[CHIRP_LINE_MAX];
	INT64_T fields;

	memset(buf,0,sizeof(*buf));

	if(!link_readline(c->link,line,CHIRP_LINE_MAX,stoptime)) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	fields = sscanf(line,"%lld %lld %lld %lld %lld %lld %lld\n",
			&buf->f_type,
			&buf->f_bsize,
			&buf->f_blocks,
			&buf->f_bfree,
			&buf->f_bavail,
			&buf->f_files,
			&buf->f_ffree);

	if(fields!=7) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	return 0;
}



static INT64_T get_result( struct chirp_client *c, time_t stoptime )
{
	char line[CHIRP_LINE_MAX];
	INT64_T result;
	INT64_T fields;

	if(!link_readline(c->link,line,CHIRP_LINE_MAX,stoptime)) {
		c->broken = 1;
		return CHIRP_ERROR_DISCONNECTED;
	}

	fields = sscanf(line,"%lld",&result);
	if(fields!=1) {
		c->broken = 1;
		return CHIRP_ERROR_DISCONNECTED;
	}

	return result;
}

INT64_T simple_command( struct chirp_client *c, time_t stoptime, char const *fmt, ... )
{
	INT64_T	result, unix_result;
	char    command[CHIRP_LINE_MAX];
	va_list args;

	va_start(args,fmt);
	vsprintf(command,fmt,args);
	va_end(args);

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	debug(D_CHIRP,"%s: %s",c->hostport,command);

	result = link_write(c->link,command,strlen(command),stoptime);
	if(result < 0) {
		c->broken = 1;
		errno = ECONNRESET;
		unix_result = -1;
	} else {
		result = get_result(c,stoptime);
		unix_result = convert_result(result);
		if(result>=0) {
			debug(D_CHIRP," = %lld",result);
		} else {
			debug(D_CHIRP," = %lld (%s)",result,strerror(errno));
		}
	}

	return unix_result;
}
